-module(wsHttp).

-behavior(gen_srv).

-include_lib("eNet/include/eNet.hrl").

-include("wsCom.hrl").

-export([
   start_link/1
   , sendResponse/5
   , send_file/5
   %% Exported for looping with a fully-qualified module name
   , chunk_loop/1
   , split_args/1
   , closeOrKeepAlive/2
]).

%% eNet callback
-export([newCon/2]).

-export([
   init/1
   , handleCall/3
   , handleCast/2
   , handleInfo/2
   , terminate/2
   , code_change/3
]).

-define(SERVER, ?MODULE).

newCon(_Sock, SupPid) ->
   supervisor:start_link(SupPid, []).

%% ********************************************  API *******************************************************************
start_link(WsMod) ->
   gen_srv:start_link(?MODULE, WsMod, []).

%% ********************************************  callback **************************************************************
init(WsMod) ->
   {ok, #wsState{wsMod = WsMod}}.

handleCall(_Msg, _State, _FROM) ->
   ?wsErr("~p call receive unexpect msg ~p ~n ", [?MODULE, _Msg]),
   {reply, ok}.

%% 默认匹配
handleCast(_Msg, _State) ->
   ?wsErr("~p cast receive unexpect msg ~p ~n ", [?MODULE, _Msg]),
   kpS.

handleInfo({tcp, _Socket, Data}, State) ->
   #wsState{stage = Stage, buffer = Buffer, socket = Socket} = State,
   case wsHttpProtocol:request(Stage, <<Buffer/binary, Data/binary>>, State) of
      {ok, NewState} ->
         {noreply, NewState};
      {done, NewState} ->
         Response = doHandle(NewState),
         #wsState{buffer = Buffer, socket = Socket, temHeader = TemHeader, method = Method} = NewState,
         case doResponse(Response, Socket, TemHeader, Method) of
            keep_alive ->
               handleInfo({tcp, _Socket, Buffer}, newWsState(NewState));
            {close, _} ->
               wsNet:close(Socket),
               ok
         end;
      Err ->
         ?wsErr("recv the http data error ~p~n", [Err]),
         send_bad_request(Socket),
         gen_tcp:close(Socket)
   end;
handleInfo({tcp_closed, _Socket}, _State) ->
   ok;

handleInfo({tcp_error, Socket, Reason}, _State) ->
   ?wsErr("the http socket error ~p~n", [Reason]),
   gen_tcp:close(Socket),
   kpS;

handleInfo({ssl, Socket, Data}, State) ->
   #wsState{stage = Stage, buffer = Buffer, socket = Socket} = State,
   case wsHttpProtocol:request(Stage, <<Buffer/binary, Data/binary>>, State) of
      {ok, NewState} ->
         {noreply, NewState};
      {done, NewState} ->
         ok;
      Err ->
         ?wsErr("recv the http data error ~p~n", [Err]),
         send_bad_request(Socket),
         ssl:close(Socket)
   end;
handleInfo({ssl_closed, Socket}, _State) ->
   ok;

handleInfo({ssl_error, Socket, Reason}, _State) ->
   ?wsErr("the http socket error ~p~n", [Reason]),
   ssl:close(Socket),
   kpS;
handleInfo({?mSockReady, Sock}, _State) ->
   inet:setopts(Sock, [{packet, raw}, {active, true}]),
   {ok, #wsState{socket = Sock}};
handleInfo({?mSockReady, Sock, SslOpts, SslHSTet}, State) ->
   case ntSslAcceptor:handshake(Sock, SslOpts, SslHSTet) of
      {ok, SslSock} ->
         ssl:setopts(Sock, [{packet, raw}, {active, true}]),
         {ok, State#wsState{socket = SslSock, isSsl = true}};
      _Err ->
         ?wsErr("ssl handshake error ~p~n", [_Err]),
         {stop, _Err, State}
   end;
handleInfo(_Msg, _State) ->
   ?wsErr("~p info receive unexpect msg ~p ~n ", [?MODULE, _Msg]),
   kpS.

terminate(_Reason, _State) ->
   ok.

code_change(_OldVsn, State, _Extra) ->
   {ok, State}.

newWsState(WsState) ->
   WsState#wsState{
      stage = reqLine
      , buffer = <<>>
      , wsReq = undefined
      , headerCnt = 0
      , temHeader = []
      , contentLength = undefined
      , temChunked = <<>>
   }.

%% @doc Execute the user callback, translating failure into a proper response.
doHandle(State) ->
   #wsState{wsMod = WsMod, method = Method, path = Path, wsReq = WsReq} = State,
   try WsMod:handle(Method, Path, WsReq) of
      %% {ok,...{file,...}}
      {ok, Headers, {file, Filename}} ->
         {file, 200, Headers, Filename, []};
      {ok, Headers, {file, Filename, Range}} ->
         {file, 200, Headers, Filename, Range};
      %% ok simple
      {ok, Headers, Body} -> {response, 200, Headers, Body};
      {ok, Body} -> {response, 200, [], Body};
      %% Chunk
      {chunk, Headers} -> {chunk, Headers, <<"">>};
      {chunk, Headers, Initial} -> {chunk, Headers, Initial};
      %% File
      {HttpCode, Headers, {file, Filename}} ->
         {file, HttpCode, Headers, Filename, {0, 0}};
      {HttpCode, Headers, {file, Filename, Range}} ->
         {file, HttpCode, Headers, Filename, Range};
      %% Simple
      {HttpCode, Headers, Body} -> {response, HttpCode, Headers, Body};
      {HttpCode, Body} -> {response, HttpCode, [], Body};
      %% Unexpected
      Unexpected ->
         ?wsErr("handle return error ~p ~p~n", [WsReq, Unexpected]),
         {response, 500, [], <<"Internal server error">>}
   catch
      throw:{ResponseCode, Headers, Body} when is_integer(ResponseCode) ->
         {response, ResponseCode, Headers, Body};
      throw:Exc:Stacktrace ->
         ?wsErr("handle catch throw ~p ~p ~p~n", [WsReq, Exc, Stacktrace]),
         {response, 500, [], <<"Internal server error">>};
      error:Error:Stacktrace ->
         ?wsErr("handle catch error ~p ~p ~p~n", [WsReq, Error, Stacktrace]),
         {response, 500, [], <<"Internal server error">>};
      exit:Exit:Stacktrace ->
         ?wsErr("handle catch exit ~p ~p ~p~n", [WsReq, Exit, Stacktrace]),
         {response, 500, [], <<"Internal server error">>}
   end.

doResponse({response, Code, UserHeaders, Body}, Socket, TemHeader, Method) ->
   Headers = [connection(UserHeaders, TemHeader), contentLength(UserHeaders, Body) | UserHeaders],
   sendResponse(Socket, Method, Code, Headers, Body),
   closeOrKeepAlive(UserHeaders, TemHeader);
doResponse({chunk, UserHeaders, Initial}, Socket, TemHeader, Method) ->
   ResponseHeaders = [{?TRANSFER_ENCODING_HEADER, <<"chunked">>}, connection(UserHeaders, TemHeader) | UserHeaders],
   sendResponse(Socket, Method, 200, ResponseHeaders, <<>>),
   Initial =:= <<"">> orelse send_chunk(Socket, Initial),
   case start_chunk_loop(Socket) of
      {error, client_closed} -> client;
      ok -> server
   end,
   close;
doResponse({file, ResponseCode, UserHeaders, Filename, Range}, Socket, TemHeader, Method) ->
   ResponseHeaders = [connection(UserHeaders, TemHeader) | UserHeaders],
   case wsUtil:file_size(Filename) of
      {error, _FileError} ->
         send_server_error(Socket),
         wsNet:close(Socket),
         exit(normal);
      Size ->
         case wsUtil:normalize_range(Range, Size) of
            undefined ->
               send_file(Socket, ResponseCode, [{<<"Content-Length">>, Size} | ResponseHeaders], Filename, {0, 0});
            {Offset, Length} ->
               ERange = wsUtil:encode_range({Offset, Length}, Size),
               send_file(Socket, 206,
                  lists:append(ResponseHeaders, [{<<"Content-Length">>, Length}, {<<"Content-Range">>, ERange}]),
                  Filename, {Offset, Length});
            invalid_range ->
               ERange = wsUtil:encode_range(invalid_range, Size),
               sendResponse(Socket, Method, 416, lists:append(ResponseHeaders, [{<<"Content-Length">>, 0}, {<<"Content-Range">>, ERange}]), <<>>)
         end,
         closeOrKeepAlive(UserHeaders, TemHeader)
   end.

%% @doc Generate a HTTP response and send it to the client.
sendResponse(Socket, Method, Code, Headers, UserBody) ->
   Body =
      case Method of
         'HEAD' ->
            <<>>;
         _ ->
            case Code of
               304 ->
                  <<>>;
               204 ->
                  <<>>;
               _ ->
                  UserBody
            end
      end,

   Response = http_response(Code, Headers, Body),

   case wsNet:send(Socket, Response) of
      ok ->
         ok;
      _Err ->
         ?wsErr("send_response error ~p~n", [_Err])
   end.


%% @doc Send a HTTP response to the client where the body is the
%% contents of the given file. Assumes correctly set response code
%% and headers.
-spec send_file(Req, Code, Headers, Filename, Range) -> ok when
   Req :: elli:wsReq(),
   Code :: elli:response_code(),
   Headers :: elli:headers(),
   Filename :: file:filename(),
   Range :: wsUtil:range().
send_file(Socket, Code, Headers, Filename, Range) ->
   ResponseHeaders = assemble_response_headers(Code, Headers),

   case file:open(Filename, [read, raw, binary]) of
      {ok, Fd} -> do_send_file(Fd, Range, Socket, ResponseHeaders);
      {error, FileError} ->
         send_server_error(Socket),
         wsNet:close(Socket),
         exit(normal)
   end,
   ok.

do_send_file(Fd, {Offset, Length}, Socket, Headers) ->
   try wsNet:send(Socket, Headers) of
      ok ->
         case wsNet:sendfile(Fd, Socket, Offset, Length, []) of
            {ok, BytesSent} -> s(file, BytesSent), ok;
            {error, Closed} when Closed =:= closed orelse Closed =:= enotconn ->
               ?wsErr("send file error")
         end;
      {error, Closed} when Closed =:= closed orelse Closed =:= enotconn ->
         ?wsErr("send file error")
   after
      file:close(Fd)
   end.

%% @doc To send a response, we must first have received everything the
%% client is sending. If this is not the case, {@link send_bad_request/1}
%% might reset the client connection.
send_bad_request(Socket) ->
   send_rescue_response(Socket, 400, <<"Bad Request">>).

send_server_error(Socket) ->
   send_rescue_response(Socket, 500, <<"Server Error">>).

send_rescue_response(Socket, Code, Body) ->
   Response = http_response(Code, Body),
   wsNet:send(Socket, Response).


%%
%% CHUNKED-TRANSFER
%%


%% @doc The chunk loop is an intermediary between the socket and the
%% user. We forward anything the user sends until the user sends an
%% empty response, which signals that the connection should be
%% closed. When the client closes the socket, the loop exits.
start_chunk_loop(Socket) ->
   %% Set the socket to active so we receive the tcp_closed message
   %% if the client closes the connection
   wsNet:setopts(Socket, [{active, once}]),
   ?MODULE:chunk_loop(Socket).

chunk_loop(Socket) ->
   {_SockType, InnerSocket} = Socket,
   receive
      {tcp_closed, InnerSocket} ->
         {error, client_closed};

      {chunk, close} ->
         case wsNet:send(Socket, <<"0\r\n\r\n">>) of
            ok ->
               wsNet:close(Socket),
               ok;
            {error, Closed} when Closed =:= closed orelse
               Closed =:= enotconn ->
               {error, client_closed}
         end;
      {chunk, close, From} ->
         case wsNet:send(Socket, <<"0\r\n\r\n">>) of
            ok ->
               wsNet:close(Socket),
               From ! {self(), ok},
               ok;
            {error, Closed} when Closed =:= closed orelse
               Closed =:= enotconn ->
               From ! {self(), {error, closed}},
               ok
         end;

      {chunk, Data} ->
         send_chunk(Socket, Data),
         ?MODULE:chunk_loop(Socket);
      {chunk, Data, From} ->
         case send_chunk(Socket, Data) of
            ok ->
               From ! {self(), ok};
            {error, Closed} when Closed =:= closed orelse
               Closed =:= enotconn ->
               From ! {self(), {error, closed}}
         end,
         ?MODULE:chunk_loop(Socket)
   after 10000 ->
      ?MODULE:chunk_loop(Socket)
   end.


send_chunk(Socket, Data) ->
   case iolist_size(Data) of
      0 -> ok;
      Size ->
         Response = [integer_to_list(Size, 16),
            <<"\r\n">>, Data, <<"\r\n">>],
         s(chunks, iolist_size(Response)),
         wsNet:send(Socket, Response)
   end.

maybe_send_continue(Socket, Headers) ->
   % According to RFC2616 section 8.2.3 an origin server must respond with
   % either a "100 Continue" or a final response code when the client
   % headers contains "Expect:100-continue"
   case lists:keyfind(?EXPECT_HEADER, 1, Headers) of
      <<"100-continue">> ->
         Response = http_response(100),
         wsNet:send(Socket, Response);
      _Other ->
         ok
   end.

http_response(Code) ->
   http_response(Code, <<>>).

http_response(Code, Body) ->
   http_response(Code, [{?CONTENT_LENGTH_HEADER, size(Body)}], Body).

http_response(Code, Headers, Body) ->
   [<<"HTTP/1.1 ">>, status(Code), <<"\r\n">>, spellHeaders(Headers), <<"\r\n">>, Body].

assemble_response_headers(Code, Headers) ->
   ResponseHeaders = http_response(Code, Headers, <<>>),
   s(resp_headers, iolist_size(ResponseHeaders)),
   ResponseHeaders.

spellHeaders(Headers) ->
   <<<<(toBinStr(Key))/binary, ": ", (toBinStr(Value))/binary, "\r\n">> || {Key, Value} <- Headers>>.

toBinStr(V) when is_integer(V) -> integer_to_binary(V);
toBinStr(V) when is_binary(V) -> V;
toBinStr(V) when is_list(V) -> list_to_binary(V);
toBinStr(V) when is_atom(V) -> atom_to_binary(V).

closeOrKeepAlive(UserHeaders, ReqHeader) ->
   case lists:keyfind(?CONNECTION_HEADER, 1, UserHeaders) of
      {_, <<"close">>} ->
         case lists:keyfind(?CONNECTION_HEADER, 1, ReqHeader) of
            {_, <<"close">>} ->
               close;
            _ ->
               keep_alive
         end;
      _ ->
         keep_alive
   end.

connection(UserHeaders, ReqHeader) ->
   case lists:keyfind(?CONNECTION_HEADER, 1, UserHeaders) of
      false ->
         case lists:keyfind(?CONNECTION_HEADER, 1, ReqHeader) of
            {_, <<"close">>} ->
               {?CONNECTION_HEADER, <<"Close">>};
            {_, <<"Close">>} ->
               {?CONNECTION_HEADER, <<"Close">>};
            _ ->
               {?CONNECTION_HEADER, <<"Keep-Alive">>}
         end;
      _ ->
         []
   end.

contentLength(Headers, Body) ->
   case lists:keyfind(?CONTENT_LENGTH_HEADER, Headers) of
      false ->
         [];
      _ ->
         {?CONTENT_LENGTH_HEADER, iolist_size(Body)}
   end.

%% @doc Split the URL arguments into a proplist.
%% Lifted from `cowboy_http:x_www_form_urlencoded/2'.
-spec split_args(binary()) -> list({binary(), binary() | true}).
split_args(<<>>) ->
   [];
split_args(Qs) ->
   Tokens = binary:split(Qs, <<"&">>, [global, trim]),
   [case binary:split(Token, <<"=">>) of
       [Token] -> {Token, true};
       [Name, Value] -> {Name, Value}
    end || Token <- Tokens].

%%
%% SIZE HELPERS
%%

%% @doc stores response part size in bytes
s(chunks, Size) ->
   case get({size, chunks}) of
      undefined ->
         put({size, chunks}, Size);
      Sum ->
         put({size, chunks}, Size + Sum)
   end;
s(Key, Size) ->
   put({size, Key}, Size).

get_sizes({{size, Key}, Value}) ->
   erase({size, Key}),
   {true, {Key, Value}};
get_sizes(_) ->
   false.

%%
%% HTTP STATUS CODES
%%

%% @doc Response code string. Lifted from `cowboy_http_req.erl'.
status(100) -> <<"100 Continue">>;
status(101) -> <<"101 Switching Protocols">>;
status(102) -> <<"102 Processing">>;
status(200) -> <<"200 OK">>;
status(201) -> <<"201 Created">>;
status(202) -> <<"202 Accepted">>;
status(203) -> <<"203 Non-Authoritative Information">>;
status(204) -> <<"204 No Content">>;
status(205) -> <<"205 Reset Content">>;
status(206) -> <<"206 Partial Content">>;
status(207) -> <<"207 Multi-Status">>;
status(226) -> <<"226 IM Used">>;
status(300) -> <<"300 Multiple Choices">>;
status(301) -> <<"301 Moved Permanently">>;
status(302) -> <<"302 Found">>;
status(303) -> <<"303 See Other">>;
status(304) -> <<"304 Not Modified">>;
status(305) -> <<"305 Use Proxy">>;
status(306) -> <<"306 Switch Proxy">>;
status(307) -> <<"307 Temporary Redirect">>;
status(400) -> <<"400 Bad Request">>;
status(401) -> <<"401 Unauthorized">>;
status(402) -> <<"402 Payment Required">>;
status(403) -> <<"403 Forbidden">>;
status(404) -> <<"404 Not Found">>;
status(405) -> <<"405 Method Not Allowed">>;
status(406) -> <<"406 Not Acceptable">>;
status(407) -> <<"407 Proxy Authentication Required">>;
status(408) -> <<"408 Request Timeout">>;
status(409) -> <<"409 Conflict">>;
status(410) -> <<"410 Gone">>;
status(411) -> <<"411 Length Required">>;
status(412) -> <<"412 Precondition Failed">>;
status(413) -> <<"413 Request Entity Too Large">>;
status(414) -> <<"414 Request-URI Too Long">>;
status(415) -> <<"415 Unsupported Media Type">>;
status(416) -> <<"416 Requested Range Not Satisfiable">>;
status(417) -> <<"417 Expectation Failed">>;
status(418) -> <<"418 I'm a teapot">>;
status(422) -> <<"422 Unprocessable Entity">>;
status(423) -> <<"423 Locked">>;
status(424) -> <<"424 Failed Dependency">>;
status(425) -> <<"425 Unordered Collection">>;
status(426) -> <<"426 Upgrade Required">>;
status(428) -> <<"428 Precondition Required">>;
status(429) -> <<"429 Too Many Requests">>;
status(431) -> <<"431 Request Header Fields Too Large">>;
status(500) -> <<"500 Internal Server Error">>;
status(501) -> <<"501 Not Implemented">>;
status(502) -> <<"502 Bad Gateway">>;
status(503) -> <<"503 Service Unavailable">>;
status(504) -> <<"504 Gateway Timeout">>;
status(505) -> <<"505 HTTP Version Not Supported">>;
status(506) -> <<"506 Variant Also Negotiates">>;
status(507) -> <<"507 Insufficient Storage">>;
status(510) -> <<"510 Not Extended">>;
status(511) -> <<"511 Network Authentication Required">>;
status(I) when is_integer(I), I >= 100, I < 1000 -> <<(integer_to_binary(I))/binary, "Status">>;
status(B) when is_binary(B) -> B.
